import numpy as np

prefix = ["T", "G", "M", "k", "", "m", "\mu ", "n", "p", "f", "a", "z", "y"]

def round(value, sig_figs=3):
    rounded_value = np.round(value, -int(np.floor(np.log10(np.abs(value))) - (sig_figs - 1)))
    if int(rounded_value) == rounded_value:
        return int(rounded_value)
    else:
        return np.round(value, -int(np.floor(np.log10(np.abs(value))) - (sig_figs - 1)))

def unit_prefix(value):
    if value < 0:
        factor = -1
    else:
        factor = 1

    value = np.abs(value)

    if value == 0:
        return 0, ""
    p = 4
    while value < 1:
        value *= 1000
        p += 1
    while value >= 1000:
        value /= 1000
        p -= 1
    
    return factor * value, prefix[p]

def unit(value, unit):
    value, p = unit_prefix(value)
    rounded_value = round(value)
    value = f"{rounded_value:.{3 - int(np.floor(np.log10(np.abs(rounded_value)))) - 1}f}"
    return f"{value} $[{p}{unit}]$"

def unit_uncertainty(value, uncertainty, unit):
    value, p = unit_prefix(value)
    uncertainty, p2 = unit_prefix(uncertainty)
    rounded_value = round(value)
    rounded_uncertainty = round(uncertainty)
    value = f"{rounded_value:.{3 - int(np.floor(np.log10(np.abs(rounded_value)))) - 1}f}"
    uncertainty = f"{rounded_uncertainty:.{3 - int(np.floor(np.log10(np.abs(rounded_uncertainty)))) - 1}f}"
    return f"{value} $\pm$ {uncertainty} $[{p}{unit} \pm {p2}{unit}]$"



import numpy as np
import cv2
import glob
import pickle


def calibrate(image_path, checkerboard_dim, save_path=None, visualize=False):
    """
    Calibrates the camera using a set of images of a checkerboard pattern.

    Args:
        image_path (str): The path to the directory containing the checkerboard images.
        checkerboard_dim (tuple): The dimensions of the checkerboard pattern (rows, columns).
        save_path (str, optional): The path to save the camera calibration parameters. If not provided, the parameters will be saved to 'camera_calibration.p' in the current directory. Defaults to None.
        visualize (bool, optional): Whether to visualize the detected corners on the images. Defaults to False.

    Returns:
        tuple: A tuple containing the camera matrix (mtx) and distortion coefficients (dist).
    """
    
    assert isinstance(image_path, str), 'image_path must be a string.'
    assert isinstance(checkerboard_dim, tuple), 'checkerboard_dim must be a tuple.'
    assert len(checkerboard_dim) == 2, 'checkerboard_dim must be a tuple of length 2.'
    assert isinstance(checkerboard_dim[0], int) and isinstance(checkerboard_dim[1], int), 'checkerboard_dim must contain integers.'
    assert isinstance(save_path, (str, type(None))), 'save_path must be a string or None.'
    assert isinstance(visualize, bool), 'visualize must be a boolean.'
    
    objp = np.zeros((checkerboard_dim[0]*checkerboard_dim[1],3), np.float32)
    objp[:,:2] = np.mgrid[0:checkerboard_dim[0],0:checkerboard_dim[1]].T.reshape(-1,2)
    
    objpoints = [] # 3d points in real world space
    imgpoints = [] # 2d points in image plane.
    
    images = glob.glob(image_path)
    for idx, fname in enumerate(images):
        img = cv2.imread(fname)
        print(fname)
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        ret, corners = cv2.findChessboardCorners(gray, checkerboard_dim, None)
        if ret == True:
            objpoints.append(objp)
            imgpoints.append(corners)
            if visualize:
                cv2.drawChessboardCorners(img, checkerboard_dim, corners, ret)
                cv2.imshow('img', img)
                cv2.waitKey(500)
    cv2.destroyAllWindows()
    
    img = cv2.imread(images[0])
    img_size = (img.shape[1], img.shape[0])
    ret, mtx, dist, rvecs, tvecs = cv2.calibrateCamera(objpoints, imgpoints, img_size, None, None)

    if save_path is not None:
        with open(save_path, 'wb') as f:
            pickle.dump([mtx, dist], f)
    else:
        with open('camera_calibration.p', 'wb') as f:
            pickle.dump([mtx, dist], f)
            
    return mtx, dist



import numpy as np
import matplotlib.pyplot as plt 
from .histogram import Histogram
from iminuit import Minuit
from iminuit.cost import LeastSquares
from ..utils.unit import unit, unit_uncertainty
from .histogram import bisection
from scipy.interpolate import PchipInterpolator
from tqdm import tqdm
import concurrent.futures
from numba import njit


@njit
def model(x, dt, risetime, decaytime, A, c):
    return A * (np.exp(-(x - dt) / decaytime) - np.exp(-(x - dt) / risetime)) + c


class Signal(Histogram):
    def __init__(self, sample, bin_width=None, confidence_level=0.95, bins=None):
        super().__init__(sample, bin_width, confidence_level, bins)
        self.sample_size = len(self.sample)
    
    def fit_worker(self):
        lsq = LeastSquares(self.x, self.y, 0.1 * np.amax(self.y), model)
        m = Minuit(lsq, 
                   dt=np.random.uniform(np.min(self.x), 0.01 * np.max(self.x)),
                   risetime=self.risetime * np.random.normal(1, 0.1),
                   decaytime=self.decaytime * np.random.normal(1, 0.1),
                   A=np.random.normal(1, 0.1) * np.max(self.sample),
                   c=np.random.normal(1, 0.1) * np.min(self.sample))
        
        m.limits["dt"] = (0, None)
        m.limits["risetime"] = (0, None)
        m.limits["decaytime"] = (0, None)
        m.limits["A"] = (0, None)

        m.migrad()
        m.hesse()
        
        return m
    
    def fit(self, risetime, decaytime):
        self.risetime = risetime
        self.decaytime = decaytime
            
        num_tasks = 100
        
        with concurrent.futures.ProcessPoolExecutor() as executor:
            results = []
            futures = [executor.submit(self.fit_worker) for _ in range(num_tasks)]

            for future in tqdm(concurrent.futures.as_completed(futures), total=num_tasks):
                results.append(future.result())
                
        chi2 = [m.fval / (len(self.x) - len(m.values)) for m in results]
        
        m = results[np.argmin(chi2)]
        
        self.dt = m.values['dt']
        self.risetime = m.values['risetime']
        self.decaytime = m.values['decaytime']
        self.A = m.values['A']
        self.c = m.values['c']
        self.dt_error = m.errors['dt']
        self.risetime_error = m.errors['risetime']
        self.decaytime_error = m.errors['decaytime']
        self.A_error = m.errors['A']
        self.c_error = m.errors['c']
        
        lsq = LeastSquares(self.x, self.y, self.e, model)
        m = Minuit(lsq, dt=self.dt, risetime=self.risetime, decaytime=self.decaytime, A=self.A, c=self.c)
        m.hesse()
        self.chi2 = m.fval / (len(self.x) - len(m.values))
        self.ndof = len(self.x) - len(m.values)

        self.y_fit = model(self.x, self.dt, self.risetime, self.decaytime, self.A, self.c)
        
        pass
    
    def fit_plot(self, units="", xlabel=""):
        plt.rcParams["font.family"] = "monospace"
        plt.rcParams["font.sans-serif"] = ["DejaVu Sans", "Arial", "Liberation Sans"]


        fit_info = [
            f'$\chi^2$\t      ${self.chi2:.3f}$\n',
            f'Drift Time \t{unit_uncertainty(self.dt, self.dt_error, units)}',
            f'Rise Time \t{unit_uncertainty(self.risetime, self.risetime_error, units)}',
            f'Decay Time \t{unit_uncertainty(self.decaytime, self.decaytime_error, units)}',
            f'Amplitude \t{unit_uncertainty(self.A, self.A_error, units)}',
            f'Constant \t{unit_uncertainty(self.c, self.c_error, units)}\n',
            f'FWHM \t\t{unit(self.fit_FWHM(1E-12), units)}',
            f'Sample Size \t{int(self.sample_size)} $[Events]$\n',
        ]
        
        plt.errorbar(self.x, self.y, yerr=self.e, fmt='o', label='Data', zorder=1, color='black')
        x = np.linspace(np.amin(self.x), np.amax(self.x), 1000)
        plt.plot(x, model(x, self.dt, self.risetime, self.decaytime, self.A, self.c), label='Fit', zorder=2, color='black', linestyle='--', linewidth=2)
        plt.legend(title='\n'.join(fit_info), bbox_to_anchor=(1.11, 1.14), loc='upper right', borderaxespad=0.)
        plt.ylabel('Probability Density')
        plt.xlabel(xlabel)
        plt.grid(True)


    def fit_FWHM(self, resolution):
        y = PchipInterpolator(self.x, self.y_fit)

        max_value = np.amax(self.y)
        half_max = max_value / 2

        peak_x = self.x[np.argmax(self.y)]

        min_x = np.amin(self.x)
        max_x = np.amax(self.x)

        left = bisection(self.x, y, half_max, resolution, min_x, peak_x)
        right = bisection(self.x, y, half_max, resolution, peak_x, max_x)

        return np.abs(right - left)
   





from .histogram import Histogram
from .gaussian import Gaussian
from .ctr import CTR
from .landau import Landau
from .signal import Signal

__all__ = ['Histogram', 'Gaussian', 'CTR', "Landau", 'Signal']




import numpy as np
import scipy.stats as stats
from scipy.interpolate import PchipInterpolator
import matplotlib.pyplot as plt


class Histogram:
    def __init__(self, sample, bin_width=None, confidence_level=0.95, bins=None, range=None):

        assert len(sample) > 0
        assert confidence_level > 0
        assert confidence_level < 1
        if range is not None:
            assert type(range) == tuple
            assert len(range) == 2
            assert range[0] < range[1]

        if bin_width is None and bins is None:
            raise ValueError("Either bin_width or bins must be specified")
        elif bin_width is not None and bins is not None:
            raise ValueError("Only one of bin_width or bins must be specified")
        elif bin_width is not None:
            pass
        else:
            bin_width = (np.amax(sample) - np.amin(sample)) / bins

        if range is not None:
            sample = sample[(sample > range[0]) & (sample < range[1])]

        self.sample = sample
        self.range = range
        self.bin_width = bin_width
        self.confidence_level = confidence_level
        if range is not None:
            self.bins = int((range[1] - range[0]) // bin_width + 1)
        else:
            self.bins = int((np.amax(sample) - np.amin(sample)) // bin_width + 1)

        if self.range is not None:
            self.__min = self.range[0] // bin_width - 1
        else:
            self.__min = np.amin(sample) // bin_width - 1

        self.__shift = self.__min * bin_width + bin_width / 2
        self.__shifted_sample = sample - self.__shift
        self.__center_adjustment = bin_width/2 - np.mean(self.__shifted_sample) % bin_width
        self.__shifted_sample += self.__center_adjustment
        self.__shift -= self.__center_adjustment

        self._x, self._y = self.__bin(self.__shifted_sample, bin_width, self.bins)
        
        self.FWHM((self.x[1] - self.x[0]) / 1E4)

    def __bin(self, sample, bin_width, bins):
        assert bin_width > 0
        assert bins > 0
        assert type(bins) == int
        assert type(bin_width) in [int, float, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64, np.float16, np.float32, np.float64]

        x = np.linspace(0, bins * bin_width, bins + 1, endpoint=True)
        y = np.zeros(bins + 1)

        for s in sample:
            index = int((s // bin_width))
            if index >= bins:  # Ensure index is within bounds
                index = bins - 1
            y[index] += 1

        return x, y

    @property
    def x(self):
        return self._x + self.__shift + self.bin_width / 2

    @x.setter
    def x(self, value):
        self._x = value - self.__shift - self.bin_width / 2

    @property
    def y(self):
        area = -np.trapezoid(self.x, self._y)

        return self._y / area
    
    @y.setter
    def y(self, value):
        self._y = value

    @property
    def e(self):
        area = -np.trapezoid(self.x, self._y)
        return self.error_estimate(self._y) / area
    
    @property
    def error_bound(self):
        lower_bound = []
        upper_bound = []
        for i in range(len(self._y)):
            lower, upper = self.poisson_confidence_interval(self._y[i])
            if not np.isfinite(lower):
                lower = 0
            if not np.isfinite(upper):
                upper = 0
            lower_bound.append(self._y[i] - lower)
            upper_bound.append(upper - self._y[i])
        return np.array([lower_bound, upper_bound]) / (np.sum(self._y) * self.bin_width)
    
    @property
    def counts(self):
        return self._y

    @property
    def counts_error(self):
        return self.error_estimate(self._y)

    def poisson_confidence_interval(self, k):
        alpha = 1 - self.confidence_level
        lower_bound = 0.5 * stats.chi2.ppf(alpha / 2, 2 * k)
        upper_bound = 0.5 * stats.chi2.ppf(1 - alpha / 2, 2 * (k + 1))

        if type(k) == np.ndarray:
            lower_bound[k == 0] = 0
        return lower_bound, upper_bound
    
    def error_estimate(self, k):
        lower_bound, upper_bound = self.poisson_confidence_interval(k)
        return (upper_bound - lower_bound) / 2

    
    def FWHM(self, resolution):
        y = PchipInterpolator(self.x, self.y)
        e_lower = PchipInterpolator(self.x, self.y - self.error_bound[0])
        e_upper = PchipInterpolator(self.x, self.y + self.error_bound[1])
        
        max_value = np.amax(self.y)
        half_max = max_value / 2

        peak_x = self.x[np.argmax(self.y)]

        min_x = np.amin(self.x)
        max_x = np.amax(self.x)

        left = bisection(self.x, y, half_max, resolution, min_x, peak_x)
        right = bisection(self.x, y, half_max, resolution, peak_x, max_x)
        
        left_gradient = (y(left + resolution) - y(left - resolution)) / (2 * resolution)
        right_gradient = (y(right + resolution) - y(right - resolution)) / (2 * resolution)
        
        left_error = e_upper(left) - e_lower(left)
        right_error = e_upper(right) - e_lower(right)
        
        left_uncertainty = left_error / left_gradient
        right_uncertainty = right_error / right_gradient
        
        uncertainty = np.sqrt(left_uncertainty ** 2 + right_uncertainty ** 2)
        
        self.dataFWHM = np.abs(right - left)
        self.dataFWHMError = uncertainty
        
        return np.abs(right - left)
    
    def plot(self, **kwargs):
        
        x = self.x if 'x' not in kwargs else kwargs.pop('x')
        
        plt.rcParams["font.family"] = "monospace"
        
        plt.errorbar(x=self.x, 
                     y=self.y,
                     yerr=self.error_bound,
                     fmt="s" if "fmt" not in kwargs else kwargs.pop("fmt"),
                     color='black' if "color" not in kwargs else kwargs.pop("color"), 
                     zorder=1 if "zorder" not in kwargs else kwargs.pop("zorder"),
                     markerfacecolor='none' if "markerfacecolor" not in kwargs else kwargs.pop("markerfacecolor"),
                     label="Data" if "label" not in kwargs else kwargs.pop("label"),
                     linestyle='none' if "linestyle" not in kwargs else kwargs.pop("linestyle"),
                     markersize=2 if "markersize" not in kwargs else kwargs.pop("markersize"))


        fill = False if "fill" not in kwargs else kwargs.pop("fill")
        
        t = np.linspace(np.amin(self.x), np.amax(self.x), 1000)

        try:
            y_lower_bound = PchipInterpolator(self.x, self.y - self.error_bound[0])(t)
        except:
            y_lower_bound = np.zeros(len(t))

        try:
            y_upper_bound = PchipInterpolator(self.x, self.y + self.error_bound[1])(t)
        except:
            y_upper_bound = np.zeros(len(t))


        if fill:
            plt.fill_between(t, y_lower_bound, y_upper_bound, color='#808080', alpha=0.25)


        plt.grid(True)
    
    def plot_unnormalized(self, **kwargs):
            
            x = self.x if 'x' not in kwargs else kwargs.pop('x')
            
            plt.rcParams["font.family"] = "monospace"
            
            plt.errorbar(x=self.x, 
                        y=self._y,
                        yerr=self.error_estimate(self._y),
                        fmt="s" if "fmt" not in kwargs else kwargs.pop("fmt"),
                        color='black' if "color" not in kwargs else kwargs.pop("color"), 
                        zorder=1 if "zorder" not in kwargs else kwargs.pop("zorder"),
                        markerfacecolor='none' if "markerfacecolor" not in kwargs else kwargs.pop("markerfacecolor"),
                        label="Data" if "label" not in kwargs else kwargs.pop("label"),
                        linestyle='none' if "linestyle" not in kwargs else kwargs.pop("linestyle"),
                        markersize=2 if "markersize" not in kwargs else kwargs.pop("markersize"))
            
            fill = False if "fill" not in kwargs else kwargs.pop("fill")
            
            t = np.linspace(np.amin(self.x), np.amax(self.x), 1000)
    
            try:
                y_lower_bound = PchipInterpolator(self.x, self._y - self.error_estimate(self._y)[0])(t)
            except:
                y_lower_bound = np.zeros(len(t))
    
            try:
                y_upper_bound = PchipInterpolator(self.x, self._y + self.error_estimate(self._y)[1])(t)
            except:
                y_upper_bound = np.zeros(len(t))
    
    
            if fill:
                plt.fill_between(t, y_lower_bound, y_upper_bound, color='#808080', alpha=0.25)
    
    
            plt.grid(True)
        
        
    def show(self):
        plt.legend()
        plt.show()
        
        
def bisection(x, y, value, tolerance, left, right):
    assert len(x) > 1
    assert tolerance > 0

    if right - left <= tolerance:
        return (left + right) / 2

    mid = (left + right) / 2

    if y(mid) < value < y(right) or y(mid) > value > y(right):
        return bisection(x, y, value, tolerance, mid, right)
    else:
        return bisection(x, y, value, tolerance, left, mid)


if __name__ == '__main__':
    sample = np.random.normal(0, 1, 10000)
    hist = Histogram(sample, 1)
    hist.plot(label='Data 1')
    hist.show()
    print(hist.FWHM(1e-5))



import cv2


def save(image, path):
    """
    Save the given image to the specified path.

    Args:
        image: The image to be saved.
        path: The path where the image will be saved.

    Returns:
        None
    """
    cv2.imwrite(path, image)
    
    
def read(path):
    """
    Reads an image from the specified path using OpenCV.

    Args:
        path (str): The path to the image file.

    Returns:
        numpy.ndarray: The image data as a NumPy array.

    """
    return cv2.imread(path)


def show(image, title='Image'):
    """
    Displays the given image in a window with the specified title.

    Parameters:
    - image: The image to be displayed.
    - title: The title of the window (default is 'Image').
    """
    cv2.imshow(title, image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    
    
    


import numpy as np
import matplotlib.pyplot as plt 
from .histogram import Histogram
from iminuit import Minuit
from iminuit.cost import LeastSquares
from ..utils.unit import unit, unit_uncertainty
from .histogram import bisection
from scipy.interpolate import PchipInterpolator
import numpy as np
import scipy.special as sp

# import uproot as ROOT

# def landau_pdf(x, loc, scale):
#     return np.array([ROOT.TMath.Landau(xi, mpv, eta, True) for xi in x])


def landau_pdf(x, loc, scale):
    """
    Calculate the Landau distribution's PDF at a given point using numerical approximation.

    Parameters:
    - x (float or array-like): The value(s) at which to evaluate the PDF.
    - loc (float): The location parameter (default is 0).
    - scale (float): The scale parameter (default is 1).

    Returns:
    - pdf (float or array-like): The value of the PDF at x.
    """
    pi_inv_sqrt = 1 / np.sqrt(np.pi)
    
    z = (x - loc) / scale
    
    pdf = pi_inv_sqrt * np.exp(-0.5 * (z + np.exp(-z))) * sp.wofz(z * 1j).real

    area = np.trapz(pdf, x)

    
    return pdf / area


def model(x, loc, scale):
    try:
        return landau_pdf(x, loc, scale)
    except:
        return np.zeros(len(x))
    
    

class Landau(Histogram):
    def __init__(self, sample, bin_width, confidence_level=0.95, bins=None, range=None):
        super().__init__(sample, bin_width, confidence_level, bins, range)
        self.sample_size = len(self.sample)
        
    def fit(self):
        min_fval = None
        m_min = None
        for _ in range(1000):
            lsq = LeastSquares(self.x, self.y, self.e, model)
            m = Minuit(lsq,
                       loc=np.random.normal(np.mean(self.sample), np.std(self.sample)),
                       scale=np.random.normal(np.std(self.sample)))
            
            m.migrad()
            m.hesse()

            if min_fval is None or m.fval < min_fval:
                min_fval = m.fval
                m_min = m
        m = m_min
        
        self.loc = m.values['loc']
        self.scale = m.values['scale']
        self.loc_error = m.errors['loc']
        self.scale_error = m.errors['scale']
        self.chi2 = m.fval / (len(self.x) - len(m.values))
        self.ndof = len(self.x) - len(m.values)

        self.y_fit = model(self.x, self.loc, self.scale)
        
    def fit_plot(self, units=""):
        plt.rcParams["font.family"] = "monospace"
        fit_info = [
            f'$\chi^2$\t      ${self.chi2:.3f}$\n',
            f'Location \t{unit_uncertainty(self.loc, self.loc_error, units)}',
            f'Scale \t\t{unit_uncertainty(self.scale, self.scale_error, units)}\n',
            f'FWHM \t\t{unit(self.fit_FWHM(1E-12), units)}\n',
            f'Sample Size \t{int(self.sample_size)} $[Events]$\n',
            f"Bin Width \t{unit(self.bin_width, units)}",
            f"Poisson C.I. \t{self.confidence_level*100:.1f} [$\%$]\n",
        ]
        
        plt.errorbar(self.x, self.y, yerr=self.e, fmt='o', label='Data', zorder=1, color='black')
        x = np.linspace(np.amin(self.x), np.amax(self.x), 1000)
        plt.plot(x, model(x, self.loc, self.scale), label='Fit', zorder=2, color='black', linestyle='--', linewidth=2)
        
        plt.legend(title='\n'.join(fit_info), bbox_to_anchor=(1.11, 1.14), loc='upper right', borderaxespad=0.)
        plt.grid(True)


    def fit_FWHM(self, resolution):
        y = PchipInterpolator(self.x, self.y_fit)

        max_value = np.amax(self.y)
        half_max = max_value / 2

        peak_x = self.x[np.argmax(self.y)]

        min_x = np.amin(self.x)
        max_x = np.amax(self.x)

        left = bisection(self.x, y, half_max, resolution, min_x, peak_x)
        right = bisection(self.x, y, half_max, resolution, peak_x, max_x)

        return np.abs(right - left)
   





import numpy as np
import matplotlib.pyplot as plt
from .histogram import Histogram
from iminuit import Minuit
from iminuit.cost import LeastSquares
from ..utils.unit import unit, unit_uncertainty
from .histogram import bisection
from scipy.interpolate import PchipInterpolator
from tqdm import tqdm
from numba import njit, prange
import concurrent.futures
from uncertainties import ufloat, nominal_value, std_dev
from uncertainties.umath import sqrt
from scipy.signal import fftconvolve
from matplotlib.ticker import AutoLocator, AutoMinorLocator



@njit
def signal(t, risetime, decaytime, lightyield):
    return (np.exp(-t/decaytime) - np.exp(-t/risetime)) / (decaytime - risetime) * lightyield

# @njit
# def signal_int(t, risetime, decaytime, lightyield):
#     x = np.linspace(0, t, 2000)
#     y = - np.trapz(signal(x, risetime, decaytime, lightyield), x)
#     return np.exp(y)

@njit
def signal_int(t, risetime, decaytime, lightyield):
    y = lightyield / (decaytime - risetime) * (risetime * np.exp(-t/risetime) - decaytime * np.exp(-t/decaytime) + decaytime - risetime)
    return np.exp(-y)


@njit
def sCTR_jit(t, risetime, decaytime, lightyield):
    return signal(t, risetime, decaytime, lightyield) * signal_int(t, risetime, decaytime, lightyield)

@njit
def sCTR(t, risetime, decaytime, lightyield):
    result = np.zeros(len(t))
    for i in range(len(t)):
        result[i] = sCTR_jit(t[i], risetime, decaytime, lightyield)
    
    return result



def model(x, dt, risetime, decaytime, lightyield):
    n = 0
    while True:
        n += 1
        val = n * np.amax(x)
        if sCTR_jit(val, risetime, decaytime, lightyield) < 1E-5:
            break

    t = np.linspace(0, n * np.amax(x), 10000)
    sctr = sCTR(t, risetime, decaytime, lightyield)

    ctr = fftconvolve(sctr, np.flip(sctr), "same")
    ct = t - np.mean(t)
    area = np.trapz(ctr, ct)


    try:
        ctr = PchipInterpolator(ct, ctr / area)(x - dt)
        ctr /= np.trapz(ctr, x)
    except:
        ctr = np.zeros(len(x))

    return ctr



class CTR(Histogram):
    def __init__(self, sample, bin_width, confidence_level=0.95, offset=True, bins=None, range=None):
        super().__init__(sample, bin_width, confidence_level, bins, range)
        self.sample_size = len(self.sample)
        
    def fit_FWHM(self, resolution):
        if not np.all(np.isfinite(self.y_fit)):
            self.fwhm = 0
            self.fwhm_error = 0
            self.fwhm_ufloat = ufloat(0, 0)
            return 0
        
        y = PchipInterpolator(self.x, self.y_fit)

        max_value = np.amax(self.y)
        half_max = max_value / 2

        peak_x = self.x[np.argmax(self.y)]

        min_x = np.amin(self.x)
        max_x = np.amax(self.x)

        left = bisection(self.x, y, half_max, resolution, min_x, peak_x)
        right = bisection(self.x, y, half_max, resolution, peak_x, max_x)

        self.risetime_ufloat = ufloat(self.risetime, self.risetime_error)
        self.decaytime_ufloat = ufloat(self.decaytime, self.decaytime_error)
        self.lightyield_ufloat = ufloat(self.lightyield, self.lightyield_error)

        factor = np.abs(right - left) / np.sqrt(self.risetime * self.decaytime / self.lightyield)
        
        
        self.fwhm_ufloat = factor * sqrt(self.risetime_ufloat * self.decaytime_ufloat / self.lightyield_ufloat)

        self.fwhm = nominal_value(self.fwhm_ufloat)
        self.fwhm_error = std_dev(self.fwhm_ufloat)

        return np.abs(right - left)

    def fit_worker(self):
        lsq = LeastSquares(self.x, self.y, self.e, model)
        m = Minuit(lsq,
                   dt=self.x[np.argmax(self.y)] * np.random.uniform(0.9, 1.1),
                   risetime=self.risetime,
                   decaytime=self.decaytime,
                   lightyield=self.lightyield * np.random.uniform(0.5, 1.5))

        m.strategy = 2
        m.limits["risetime"] = (1E-12, 10E-9)
        m.limits["decaytime"] = (20E-9, 300E-9)
        # m.limits['lightyield'] = (100, 100000)

        m.fixed['risetime'] = False
        m.fixed['decaytime'] = True
        m.fixed['lightyield'] = False

        m.simplex()
        m.migrad(iterate=100)
        try:
            m.minos()
        except:
            pass
        
        m.fixed['risetime'] = False
        m.fixed['decaytime'] = True
        m.fixed['lightyield'] = True

        m.simplex()
        m.migrad(iterate=100)
        try:
            m.minos()
        except:
            pass

        m.fixed['risetime'] = True
        m.fixed['decaytime'] = True
        m.fixed['lightyield'] = False

        m.simplex()
        m.migrad(iterate=100)

        try:
            m.minos()
        except:
            pass
        
        return m

    def fit(self, risetime, decaytime, lightyield, iterations=1, fwhm_resolution=1E-12):
        self.risetime = risetime
        self.decaytime = decaytime
        self.lightyield = lightyield


        with concurrent.futures.ProcessPoolExecutor() as executor:
            results = []
            futures = [executor.submit(self.fit_worker) for _ in range(iterations)]

            for future in tqdm(concurrent.futures.as_completed(futures), total=iterations):
                results.append(future.result())
        
        results = np.array(results)
        mask = np.array([minuit.valid for minuit in results])
        m = results[mask]
        
        chi2 = [m.fval / (len(self.x) - len(m.values)) for m in results]
        m = results[np.argmin(chi2)]

        self.dt = m.values['dt']
        self.risetime = m.values['risetime']
        self.decaytime = m.values['decaytime']
        self.lightyield = m.values['lightyield']
        self.dt_error = m.errors['dt']
        self.risetime_error = m.errors['risetime']
        self.decaytime_error = m.errors['decaytime']
        self.lightyield_error = m.errors['lightyield']

        self.chi2 = m.fval / (len(self.x) - len(m.values))
        self.ndof = len(self.x) - len(m.values)

        self.y_fit = model(self.x, self.dt, self.risetime, self.decaytime, self.lightyield)
        self.fit_FWHM(fwhm_resolution)

    def fit_plot(self, resolution=160E-12, fill=False):
        plt.rcParams["font.family"] = "monospace"
        # plt.rcParams["font.sans-serif"] = ["DejaVu Sans", "Arial", "Liberation Sans"]
        
        fit_info = [
            f'$\chi^2$\t      ${self.chi2:.3f}$\n',
            f'$\Delta t$ \t\t{unit_uncertainty(self.dt, self.dt_error, "s")}',
            f'Rise Time \t{unit_uncertainty(self.risetime, self.risetime_error, "s")}',
            f'Decay Time \t{unit_uncertainty(self.decaytime, self.decaytime_error, "s")}',
            f'Light Yield \t{unit_uncertainty(self.lightyield, self.lightyield_error, "ph")}\n',
            f'Model FWHM \t{unit_uncertainty(self.fwhm, self.fwhm_error, "s")}',
            f'Data FWHM \t{unit_uncertainty(self.dataFWHM, self.dataFWHMError, "s")}\n',
            f'Sample Size \t{int(self.sample_size)} $[Events]$',
            f"Time Scale \t{unit(resolution, 's/pt')}\n",
            f"Bin Width \t{unit(self.bin_width, 's')}",
            f"Poisson C.I. \t{self.confidence_level*100:.1f} [$\%$]\n",
        ]
      

        plt.errorbar(self.x, self.y, yerr=self.error_bound, fmt='o', label='Data', zorder=1, color='#2b2b2b')
        x = np.linspace(np.amin(self.x), np.amax(self.x), 1000)
        plt.plot(x, model(x, self.dt, self.risetime, self.decaytime, self.lightyield), label='Fit', zorder=2, color='#2b2b2b', linestyle='--',
                 linewidth=2)
        
        data_area = np.trapezoid(self.x, self.y)
        fit_area = np.trapezoid(x, model(x, self.dt, self.risetime, self.decaytime, self.lightyield))
        
        t = np.linspace(np.amin(self.x), np.amax(self.x), 1000)
        
        try:
            y_lower_bound = PchipInterpolator(self.x, self.y - self.error_bound[0])(t)
        except:
            y_lower_bound = np.zeros(len(t))

        try:
            y_upper_bound = PchipInterpolator(self.x, self.y + self.error_bound[1])(t)
        except:
            y_upper_bound = np.zeros(len(t))


        if fill:
            plt.fill_between(t, y_lower_bound, y_upper_bound, color='#a0a0a0', alpha=0.25)

        ax = plt.gca()
        ax.xaxis.set_major_locator(AutoLocator())
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_major_locator(AutoLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        ax.tick_params(direction = 'in', length = 7, top = True, right = True)
        ax.tick_params(direction = 'in', length = 4, which = 'minor', top = True, right = True)
        plt.legend(title='\n'.join(fit_info), bbox_to_anchor=(1.11, 1.14), loc='upper right', borderaxespad=0.)
        plt.xlabel('Time Difference [s]')
        plt.ylabel('Probability Density')
        plt.grid(True)











from .calibration import calibrate
from .undistort import undistort, read_calibration_data
from .io import save, read, show
from .presentation import divide, square

__all__ = ['calibrate', 'undistort', 'read_calibration_data', 'save', 'read', 'show', 'divide', 'square']



import numpy as np
import cv2 


def square(img, fov, focal_length, pixel_size):
    """
    Draws a square on the input image.

    Args:
        img (numpy.ndarray): The input image.
        fov (float): The field of view in degrees.
        focal_length (float): The focal length of the camera in mm.
        pixel_size (float): The size of each pixel in the camera sensor in mm.

    Returns:
        numpy.ndarray: The image with a square drawn on it.
    """
    assert isinstance(img, np.ndarray), 'img must be a numpy array.'
    assert isinstance(fov, (int, float)), 'fov must be an integer or float.'
    assert isinstance(focal_length, (int, float)), 'focal_length must be an integer or float.'
    assert isinstance(pixel_size, (int, float)), 'pixel_size must be an integer or float.'
    
    h, w = img.shape[:2]
    sensor_size = 2 * np.tan(np.deg2rad(fov)/2) * focal_length
    pixel_density = 1/pixel_size
    pixel = int(sensor_size * pixel_density)
    cv2.rectangle(img, (int(w/2 - pixel/2), int(h/2 - pixel/2)), (int(w/2 + pixel/2), int(h/2 + pixel/2)), (0, 255, 0), 2)  
    
    
    return img


def divide(img, fov, focal_length, pixel_size, portions):
    """
    Divides the input image into portions and draws rectangles on each portion.

    Parameters:
    img (numpy.ndarray): The input image.
    fov (float): The field of view in degrees.
    focal_length (float): The focal length of the camera in mm.
    pixel_size (float): The size of each pixel in the image.
    portions (int): The number of portions to divide the image into.

    Returns:
    numpy.ndarray: The image with rectangles drawn on each portion.
    """
    assert isinstance(img, np.ndarray), 'img must be a numpy array.'
    assert isinstance(fov, (int, float)), 'fov must be an integer or float.'
    assert isinstance(focal_length, (int, float)), 'focal_length must be an integer or float.'
    assert isinstance(pixel_size, (int, float)), 'pixel_size must be an integer or float.'
    assert isinstance(portions, int), 'portions must be an integer.'
    
    h, w = img.shape[:2]
    sensor_size = 2 * np.tan(np.deg2rad(fov)/2) * focal_length
    pixel_density = 1/pixel_size
    pixel = int(sensor_size * pixel_density)
    for i in range(portions):
        for j in range(portions):
            cv2.rectangle(img, (int(w/2 - pixel/2 + i*pixel/portions), int(h/2 - pixel/2 + j*pixel/portions)), (int(w/2 - pixel/2 + (i+1)*pixel/portions), int(h/2 - pixel/2 + (j+1)*pixel/portions)), (0, 255, 0), 2)

    return img




import cv2
import pickle


def read_calibration_data(calibration_file=None):
    """
    Read calibration data from a file.

    Args:
        calibration_file (str, optional): Path to the calibration file. If not provided, the default file name 'camera_calibration.p' will be used.

    Returns:
        tuple: A tuple containing the camera matrix (mtx) and distortion coefficients (dist).
    """
    if calibration_file is None:
        calibration_file = 'camera_calibration.p'
        
    with open(calibration_file, 'rb') as f:
        mtx, dist = pickle.load(f)
    
    return mtx, dist


def undistort(img, mtx=None, dist=None):
    """
    Undistorts an image using camera calibration parameters.

    Args:
        img (numpy.ndarray): The input image to be undistorted.
        mtx (numpy.ndarray, optional): The camera matrix. If not provided, it will be read from calibration data.
        dist (numpy.ndarray, optional): The distortion coefficients. If not provided, they will be read from calibration data.

    Returns:
        numpy.ndarray: The undistorted image.

    """
    if mtx is None or dist is None:
        mtx, dist = read_calibration_data()
        
    h,  w = img.shape[:2]
    newcameramtx, roi = cv2.getOptimalNewCameraMatrix(mtx, dist, (w,h), 1, (w,h))
    dst = cv2.undistort(img, mtx, dist, None, newcameramtx)
    x, y, w, h = roi
    dst = dst[y:y+h, x:x+w]
    
    return dst





import numpy as np
import matplotlib.pyplot as plt 
from .histogram import Histogram
from iminuit import Minuit
from iminuit.cost import LeastSquares
from ..utils.unit import unit, unit_uncertainty
from .histogram import bisection
from scipy.interpolate import PchipInterpolator
from tqdm import tqdm
from uncertainties import ufloat, nominal_value, std_dev
from uncertainties.umath import sqrt
from matplotlib.ticker import AutoLocator, AutoMinorLocator


def model(x, mu, sigma):
    return 1/(sigma * np.sqrt(2 * np.pi)) * np.exp(-0.5 * ((x - mu) / sigma) ** 2)


class Gaussian(Histogram):
    def __init__(self, sample, bin_width=None, confidence_level=0.95, bins=None, range=None):
        super().__init__(sample, bin_width, confidence_level, bins, range)
        self.sample_size = len(self.sample)
        
    def fit(self, iterations=1):
        min_fval = None
        m_min = None
        for _ in tqdm(range(iterations)):
            lsq = LeastSquares(self.x, self.y, self.e, model)
            m = Minuit(lsq, mu=np.random.normal(1, 0.1) * np.mean(self.sample), sigma=np.random.normal(1, 0.1) * np.std(self.sample))
            
            m.migrad()
            m.hesse()

            if min_fval is None or m.fval < min_fval:
                min_fval = m.fval
                m_min = m
        m = m_min
        
        self.mu = m.values['mu']
        self.sigma = m.values['sigma']
        self.mu_error = m.errors['mu']
        self.sigma_error = m.errors['sigma']
        
        self.sigma_ufloat = ufloat(self.sigma, self.sigma_error)
        
        lsq = LeastSquares(self.x, self.y, self.e, model)
        m = Minuit(lsq, mu = self.mu, sigma = self.sigma)
        m.hesse()
        self.chi2 = m.fval / (len(self.x) - len(m.values))
        self.ndof = len(self.x) - len(m.values)

        self.y_fit = model(self.x, self.mu, self.sigma)
        self.fit_FWHM(1E-12)
        
    def fit_plot(self, units="", xlabel="", timescale=None, fill=False):
        plt.rcParams["font.family"] = "monospace"
        plt.rcParams["font.sans-serif"] = ["DejaVu Sans", "Arial", "Liberation Sans"]

        if timescale is not None:
            fit_info = [
                f'$\chi^2$\t      ${self.chi2:.3f}$\n',
                f'Mean \t\t{unit_uncertainty(self.mu, self.mu_error, units)}',
                f'Std. Dev. \t{unit_uncertainty(self.sigma, self.sigma_error, units)}\n',
                f'Model FWHM \t{unit_uncertainty(self.fwhm, self.fwhm_error, "s")}',
                f'Data FWHM \t{unit_uncertainty(self.dataFWHM, self.dataFWHMError, "s")}\n',
                f'Sample Size \t{int(self.sample_size)} $[Events]$',
                f"Time Scale \t{unit(timescale, 's/pt')}\n",
                f"Bin Width \t{unit(self.bin_width, 's')}",
                f"Poisson C.I. \t{self.confidence_level*100:.1f} [$\%$]\n",
            ]
        else:
            fit_info = [
                f'$\chi^2$\t      ${self.chi2:.3f}$\n',
                f'Mean \t\t{unit_uncertainty(self.mu, self.mu_error, units)}',
                f'Std. Dev. \t{unit_uncertainty(self.sigma, self.sigma_error, units)}\n',
                f'Model FWHM \t{unit_uncertainty(self.fwhm, self.fwhm_error, "s")}',
                f'Data FWHM \t{unit_uncertainty(self.dataFWHM, self.dataFWHMError, "s")}\n',
                f'Sample Size \t{int(self.sample_size)} $[Events]$\n',
                f"Bin Width \t{unit(self.bin_width, 's')}",
                f"Poisson C.I. \t{self.confidence_level*100:.1f} [$\%$]\n",
            ]
        
        plt.errorbar(self.x, self.y, yerr=self.e, fmt='o', label='Data', zorder=1, color='black')
        x = np.linspace(np.amin(self.x), np.amax(self.x), 1000)
        plt.plot(x, model(x, self.mu, self.sigma), label='Fit', zorder=2, color='black', linestyle='--', linewidth=2)
        
        t = np.linspace(np.amin(self.x), np.amax(self.x), 1000)
        try:
            y_lower_bound = PchipInterpolator(self.x, self.y - self.error_bound[0])(t)
        except:
            y_lower_bound = np.zeros(len(t))

        try:
            y_upper_bound = PchipInterpolator(self.x, self.y + self.error_bound[1])(t)
        except:
            y_upper_bound = np.zeros(len(t))


        if fill:
            plt.fill_between(t, y_lower_bound, y_upper_bound, color='#a0a0a0', alpha=0.25)
        
        ax = plt.gca()
        ax.xaxis.set_major_locator(AutoLocator())
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_major_locator(AutoLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        ax.tick_params(direction = 'in', length = 7, top = True, right = True)
        ax.tick_params(direction = 'in', length = 4, which = 'minor', top = True, right = True)
        plt.legend(title='\n'.join(fit_info), bbox_to_anchor=(1.11, 1.14), loc='upper right', borderaxespad=0.)
        plt.ylabel('Probability Density')
        plt.xlabel(xlabel)
        plt.grid(True)


    def fit_FWHM(self, resolution):
        y = PchipInterpolator(self.x, self.y_fit)

        max_value = np.amax(self.y)
        half_max = max_value / 2

        peak_x = self.x[np.argmax(self.y)]

        min_x = np.amin(self.x)
        max_x = np.amax(self.x)

        left = bisection(self.x, y, half_max, resolution, min_x, peak_x)
        right = bisection(self.x, y, half_max, resolution, peak_x, max_x)

        self.fwhm_ufloat = 2.355 * self.sigma_ufloat
        
        self.fwhm = nominal_value(self.fwhm_ufloat)
        self.fwhm_error = std_dev(self.fwhm_ufloat)
        
        return np.abs(right - left)
   


